# gpio2

Export elegant API to control Raspberry Pi GPIO pins with Node.js. Provide both synchronous and asynchronous **Promise** API. Use interrupts to detect the state change of the GPIO pins and fire events what is much faster and more sensitive than file monitoring such as fs.watch or chokidar.

## Installation

```bash
npm install gpio2
```

## Usage

### with ES5

```
require('gpio2').Gpio;

const pinIn = 'gpio21', //or 40
      pinOut = 32;

let gpio = new Gpio(portIn);

gpio.export(Gpio.DIR_IN, {
  debounceTimeout: 500
});

gpio.on('rising', function(e){
  console.log('rising', e);
});
```

### with ES6

```
'use strict'

const Gpio = require("gpio2.js").Gpio;

const sleep = (millsec) => {
  console.log("sleep..." + millsec);
  return new Promise((resolve) => setTimeout(() => resolve(), millsec));
};

let gpio0 = new Gpio(40),
    gpio1 = new Gpio(32);

let main = async () => {
  await gpio1.export('in','falling');

  gpio1.watch(function(err, value){
    console.log('watched', gpio1.edge, gpio1.value, value);
  });

  await gpio0.export();

  while(1){
    await gpio0.toggleValue(); //make a signal changed per second.
    await sleep(1000);
  }
}

process.on("SIGINT", function(){
  gpio0.unexport();
  gpio1.unexport();

  console.log('shutdown!');
  process.exit(0);
});

main();
```

## API

### Class Gpio extends [events.EventEmitter](https://nodejs.org/dist/latest-v4.x/docs/api/events.html#events_class_eventemitter)

#### Methods

  * [Gpio(pinNumber) - Constructor](#constructor)
  * [export([options]) - Export the GPIO to userspace](#exportoptions)
  * [unexport() - Unexport the GPIO](#unexport)
  * [setValue(value) - Set GPIO value asynchronously(returns a promise)](#setvaluevalue)
  * [getValue(value) - Get GPIO value asynchronously(returns a promise)](#getvaluevalue)
  * [toggleValue() - Change GPIO value asynchronously(returns a promise)](#trogglevalue)

#### Properties

  * [value:0|1 - Set/Get GPIO value synchronously](#value01)
  * [direction:string - The input/output direction of a GPIO pin](#directionstring)
  * [activeLow:boolean - GPIO activeLow setting](#activelowboolean)

#### Events

  * [rising - Triggering an event when GPIO signal `rising`](#eventrising)
  * [falling - Triggering an event when GPIO signal `falling`](#eventfalling)
  * [change - Triggering an event when GPIO signal `rising` or `falling`](#eventchange)

#### Constants

  - Gpio.HIGH = 1;
  - Gpio.LOW = 0;
  - Gpio.DIR_IN = 'in';
  - Gpio.DIR_OUT = 'out';
  - Gpio.EDGE_NONE = 'none';
  - Gpio.EDGE_RISING = 'rising';
  - Gpio.EDGE_FALLING = 'falling';
  - Gpio.EDGE_BOTH = 'both';

---

##### constructor

`new Gpio(pin)` creates a GPIO pin instance. The arguments `pin` can be a number (pin number) or a string starts with 'gpio' (gpio number):

```
let pin1 = new Gpio('gpio21'); //create gpio21

//equal to: let pin12 = new Gpio('gpio12');
let pin2 = new Gpio(32); //create #32 pin of Pi, which is gpio12
```

The pin number mapping to GPIO number as below:

<table>
  <tr>
    <td>+ 3.3v</td><td>1</td><td>2</td><td>+ 5v</td></tr>
  <tr>
    <td>I2C SDA / <em>GPIO 2</em></td><td>3</td><td >
      4</td><td>+ 5v</td></tr>
  <tr>
    <td>I2C SCL / <em>GPIO 3</em></td><td>5</td><td>6</td><td>Ground</td></tr>
  <tr>
    <td>Clock / <em>GPIO 4</em></td><td>7</td><td>8</td><td>TX / <em>GPIO 14</em></td></tr>
  <tr>
    <td>--</td><td>9</td><td>10</td><td>RX / <em>GPIO 15</em></td></tr>
  <tr>
    <td><em>GPIO 17</em></td><td>11</td><td>12</td><td><em>GPIO 18</em></td></tr>
  <tr>
    <td><em>GPIO 27</em></td><td>13</td><td>14</td><td>--</td></tr>
  <tr>
    <td><em>GPIO 22</em></td><td>15</td><td>16</td><td><em>GPIO 23</em></td></tr>
  <tr>
    <td>+ 3.3V</td><td>17</td><td>18</td><td><em>GPIO 24</em></td></tr>
  <tr>
    <td>SPI MOSI / <em>GPIO 10</em></td><td>19</td><td>20</td><td>--</td></tr>
  <tr>
    <td>SPI MISO / <em>GPIO 9</em></td><td>21</td><td>22</td><td><em>GPIO 25</em></td></tr>
  <tr>
    <td>SPI SCLK / <em>GPIO 11</em></td><td>23</td><td>24</td><td>SPI CE0 / <em>GPIO 8</em></td></tr>
  <tr>
    <td>--</td><td>25</td><td>26</td><td>SPI CE1 / <em>GPIO 7</em></td></tr>
  <tr>
    <td colspan="4">Model A+ and Model B+ additional pins</td></tr>
  <tr>
    <td>ID_SD</td><td>27</td><td>28</td><td>ID_SC</td></tr>
  <tr>
    <td><em>GPIO 5</em></td><td>29</td><td>30</td><td>--</td></tr>
  <tr>
    <td><em>GPIO 6</em></td><td>31</td><td>32</td><td><em>GPIO 12</em></td></tr>
  <tr>
    <td><em>GPIO 13</em></td><td>33</td><td>34</td><td>--</td></tr>
  <tr>
    <td><em>GPIO 19</em></td><td>35</td><td>36</td><td><em>GPIO 16</em></td></tr>
  <tr>
    <td><em>GPIO 26</em></td><td>37</td><td>38</td><td><em>GPIO 20</em></td></tr>
  <tr>
    <td>--</td><td>39</td><td>40</td><td><em>GPIO 21</em></td></tr>
</table>

##### export([options])

Exports a GPIO to userspace.

- [options: object]  Additional options. 

	**The options argument supports the following:**

	- direction: A string specifying whether the GPIO should be configured as an input or output. The valid values are: 'in', 'out'..
	
	- activeLow: Specifies whether the values read from or written to the GPIO should be inverted. The interrupt generating edge for the GPIO also follow this this setting. The valid values for activeLow are true and false. Setting activeLow to true inverts. The default value is false.
	
	- debounceTimeout: Can be used to software debounce a button or switch using a timeout. Specified in milliseconds. The default value is 0.  

##### unexport()

Unexports a GPIO from userspace and release all resources.

##### setValue(value)

Set a velue to a GPIO **asynchronously**. Returns a promise.

##### getValue(value) 

Set a velue to a GPIO **asynchronously**. Returns a promise.

##### toggleValue()

Change GPIO value **asynchronously**. Returns a promise.

```js
async function run(gpio){
    while(1){
        gpio.toggleValue(); //blink gpio value every 0.5 second to send a signal.
        sleep(500);
    }
}
```

##### value:0|1

Get or set a velue to a GPIO **synchronously**.

##### direction:string

The pin direction, pass either Gpio.DIR_IN for read mode or Gpio.DIR_OUT for write mode. Defaults to DIR_OUT.


##### activeLow: boolean

Specifies whether the values read from or written to the GPIO should be inverted. The interrupt generating edge for the GPIO also follow this this setting. The valid values for activeLow are true and false. Setting activeLow to true inverts. The default value is false.

---

#### events

If a pin is in direction of Gpio.DIR_IN. Three type of events can be fired when needed.

##### event:rising

When register listener to `rising` event [the rising interrupt edges](https://www.raspberrypi.org/documentation/hardware/raspberrypi/gpio/README.md) should be configured implictly and the GPIO will trigger the `rising` event.

```js
gpio.on('rising', function(){
	console.log('A rising signal detected!');
});
```

##### event:falling

When register listener to `falling` event [the falling interrupt edges](https://www.raspberrypi.org/documentation/hardware/raspberrypi/gpio/README.md) should be configured implictly and the GPIO will trigger the `falling` event.

##### event:change

When register listener to `change` event [the both(rising and falling) interrupt edges](https://www.raspberrypi.org/documentation/hardware/raspberrypi/gpio/README.md) should be configured implictly and the GPIO will trigger the `change` event(on both rising and falling edges).

##### Note:

Registering events to rising and falling will implictly change the interrupt edges as well as unregistering events:

```js
let gpio = new Gpio(40);
gpio.export(Gpio.DIR_IN);
assertEqual(gpio.edge, Gpio.EDGE_NONE); 
gpio.on('rising', function(){...});
assertEqual(gpio.edge, Gpio.EDGE_RISING); 
gpio.on('falling', function(){...});
assertEqual(gpio.edge, Gpio.EDGE_BOTH);
gpio.removeListener('rising');
assertEqual(gpio.edge, Gpio.EDGE_FALLING);
gpio.removeListener('falling'); 
assertEqual(gpio.edge, Gpio.EDGE_NONE);
``` 

## Thanks

[onoff](https://github.com/fivdi/onoff) - awesome API for GPIO access. A few codes are borrow from there.

[rpi-gpio](https://github.com/JamesBarwell/rpi-gpio.js) - first library I used to control my Raspberry Pi. Inspire me to create this project.

## LICENSE

MIT

