'use strict';

var fs = require("fs"),
    Epoll = require('epoll').Epoll;

var consoler = require('consoler');

var rev = fs.readFileSync("/proc/cpuinfo").toString().split("\n").filter(function(line) {
  return line.indexOf("Revision") == 0;
})[0].split(":")[1].trim();
rev = parseInt(rev, 16) < 3 ? 1 : 2; // http://elinux.org/RPi_HardwareHistory#Board_Revision_History

var devicePath = "/sys/class/gpio";
if(!fs.existsSync(devicePath)){
  devicePath = "/sys/devices/virtual/gpio"; //fall back to pre 3.18.x kernel
}

var gpioMapping = {
  "3": 0,
  "5": 1,
  "7": 4,
  "8": 14,
  "10": 15,
  "11": 17,
  "12": 18,
  "13": 21,
  "15": 22,
  "16": 23,
  "18": 24,
  "19": 10,
  "21": 9,
  "22": 25,
  "23": 11,
  "24": 8,
  "26": 7,

  // Model A+ and Model B+ pins
  "29": 5,
  "31": 6,
  "32": 12,
  "33": 13,
  "35": 19,
  "36": 16,
  "37": 26,
  "38": 20,
  "40": 21
};

if (rev == 2) {
  gpioMapping["3"] = 2;
  gpioMapping["5"] = 3;
  gpioMapping["13"] = 27;
}

const BuffZero = new Buffer('0'),
      BuffOne = new Buffer('1');

const ErrGpioAlreadyExported = 'The gpio instance is already exported.';
const ErrGpioUnexported = 'The gpio instance is unexported.';

function errPromise(errmsg){
  consoler('error', errmsg);
  return Promise.reject(new Error(errmsg));
}

function pollerEventHandler(err, fd, events) {
  var value = this.value,
    callbacks = this._ihdl.slice(0);

  if (this.debounceTimeout > 0) {
    setTimeout(function () {
      if (this._ihdl.length > 0) {
        // Read current value before polling to prevent unauthentic interrupts.
        this.value;
        this.poller.modify(this.file.value, Epoll.EPOLLPRI | Epoll.EPOLLONESHOT);
      }
    }.bind(this), this.debounceTimeout);
  }

  callbacks.forEach(function (callback) {
    callback(err, value);
  });
}

function watch(gpio, callback){
  var events;

  gpio._ihdl.push(callback);

  if (gpio._ihdl.length === 1) {
    events = Epoll.EPOLLPRI;
    if (gpio.debounceTimeout > 0) {
      events |= Epoll.EPOLLONESHOT;
    }
    gpio.poller.add(gpio.file.value, events);
  }
}

function unwatch(gpio, callback) {
  if (gpio._ihdl.length > 0) {
    if (typeof callback !== 'function') {
      gpio._ihdl.length = 0;
    } else {
      gpio._ihdl = gpio._ihdl.filter(function (listener) {
        return callback !== listener;
      });
    }

    if (gpio._ihdl.length === 0) {
      gpio.poller.remove(gpio.file.value);
    }
  }
}

function triggerEvent(err, value){
  if(err){
    throw err;
  }else{
    this.emit(value?'rising':'falling', value);
    this.emit('change', value);
  }       
}

function updateEdge(gpio, type){
  var emitter = gpio;
  var changes = emitter.listeners('change').length;
  var risings = emitter.listeners('rising').length;
  var fallings = emitter.listeners('falling').length;

  if(type === 'change'){
    changes ++;
  }else if(type === 'rising'){
    risings ++;
  }else if(type === 'falling'){
    fallings ++;
  }

  if(changes > 0 || 
    risings > 0 && fallings > 0){
    gpio.edge = GpioExport.EDGE_BOTH;
    if(gpio._ihdl.length <= 0){
      watch(gpio, triggerEvent.bind(gpio));
    }
  }else if(risings > 0){
    gpio.edge = GpioExport.EDGE_RISING;
    if(gpio._ihdl.length <= 0){
      watch(gpio, triggerEvent.bind(gpio));
    }    
  }else if(fallings > 0){
    gpio.edge = GpioExport.EDGE_FALLING;
    if(gpio._ihdl.length <= 0){
      watch(gpio, triggerEvent.bind(gpio));
    }       
  }else{
    gpio.edge = GpioExport.EDGE_NONE;
    unwatch(gpio);
  }
}

/**
 * Constructor. 
 *
 * Thanks to: https://github.com/fivdi/onoff
 *
 * pin: number | string     // The pin number of rpi
 *
 */

function GpioExport(pin){
  var gpioNumber;

  if(/^gpio/ig.test(pin)){
    gpioNumber = parseInt(pin.slice(4));
  }else{
    gpioNumber = gpioMapping[pin];
  }

  var resRoot = devicePath + '/gpio' + gpioNumber;

  this.gpioNumber = gpioNumber;

  this.rev = rev;

  this.path = {
    root: resRoot,
    direction: resRoot + '/direction',
    value: resRoot + '/value',
    edge: resRoot + '/edge',
    activeLow: resRoot + '/active_low'
  };

  this.file = {};

  this.buffer = new Buffer(1);

  Object.defineProperty(this, '_ihdl', {
    value: [],
    writable: false,
    configurable: false,
    enumerable: false
  });

  this.on('removeListener', function(type){
    updateEdge(this);
  });

  this.on('newListener', function(type){
    updateEdge(this, type);
  });
}

GpioExport.HIGH = 1;
GpioExport.LOW = 0;
GpioExport.DIR_IN = 'in';
GpioExport.DIR_OUT = 'out';
GpioExport.EDGE_NONE = 'none';
GpioExport.EDGE_RISING = 'rising';
GpioExport.EDGE_FALLING = 'falling';
GpioExport.EDGE_BOTH = 'both';

var util = require('util'), EventEmitter = require('events').EventEmitter;
util.inherits(GpioExport, EventEmitter);

Object.assign(GpioExport.prototype, {
  /*
   * Exports a GPIO to userspace.
   *
   *
   * [options: object]        // Additional options. [optional]
   *
   * The options argument supports the following:
   *
   * activeLow: boolean       // Specifies whether the values read from or
   *                          // written to the GPIO should be inverted. The
   *                          // interrupt generating edge for the GPIO also
   *                          // follow this this setting. The valid values for
   *                          // activeLow are true and false. Setting activeLow
   *                          // to true inverts. The default value is false.
   *
   * debounceTimeout: number  // Can be used to software debounce a button or
   *                          // switch using a timeout. Specified in
   *                          // milliseconds. The default value is 0.        
   */

  export: function(direction, opts){
    direction = direction || GpioExport.DIR_OUT;
    opts = opts || {};

    if(this.file.value){
      return errPromise(ErrGpioAlreadyExported);
    }

    fs.writeFileSync(devicePath + '/export', this.gpioNumber);

    // Again, thanks to https://github.com/fivdi/onoff
    var st = Date.now(), timeout = 3000;
    while(!this.file.value && Date.now() - st <= timeout){
      try{
        // Cache value for performance.
        this.file.value = fs.openSync(this.path.value, 'r+');
      }catch(ex){}
    }

    if(this.file.value){
      // Read current value before polling to prevent unauthentic interrupts.
      this.value;
      this.poller = new Epoll(pollerEventHandler.bind(this));
      this.direction = direction;
      this.activeLow = opts.activeLow;
      this.debounceTimeout = 0|opts.debounceTimeout;

      return Promise.resolve();
    }else{
      return Promise.reject(new Error('time out'));
    }
  },
  unexport: function(){
    if(!this.file.value){
      return errPromise(ErrGpioUnexported);
    }
    
    this.removeAllListeners();

    fs.closeSync(this.file.value);
    delete this.file.value;

    fs.writeFileSync(devicePath + '/unexport', this.gpioNumber);

    return Promise.resolve();
  },

  setValue: function(value){
    var fd = this.file.value;

    if(!fd){
      return errPromise(ErrGpioUnexported);
    }
    return new Promise(function(resolve, reject){
      var buff = value ? BuffOne : BuffZero;

      fs.write(fd, buff, 0, 1, 0, function(err){
        if(!err) resolve(value);
        else reject(err);
      });
    });
  },
  getValue: function(){
    var fd = this.file.value,
        buff = this.buffer;

    if(!fd){
      return errPromise(ErrGpioUnexported);
    }
    return new Promise(function(resolve, reject){
      fs.read(fd, buff, 0, 1, 0, function(err, bytes, buf){
        if(!err) resolve(buf.toString()|0);
        else reject(err);
      });
    });
  },

  toggleValue: function(){
    var self = this;
    return this.getValue().then(function(value){
      return self.setValue(!value - 0);
    });
  }
});

Object.defineProperty(GpioExport.prototype, 'value', {
  set: function(value){
    var fd = this.file.value,
        buff = value ? BuffOne : BuffZero;

    if(fd){
      return fs.writeSync(fd, buff, 0, 1, 0);
    }
  },
  get: function(){
    var fd = this.file.value,
        buff = this.buffer;

    if(fd){
      fs.readSync(fd, buff, 0, 1, 0);
      return buff.toString()|0;
    }
  },
  configurable: false
});

Object.defineProperty(GpioExport.prototype, 'edge', {
  set: function(value){
    if(this.direction !== GpioExport.DIR_IN &&
      value !== GpioExport.EDGE_NONE){
      consoler('warning', 'The interrupt generating edges for only GPIO input. Not a GPIO output.');
      return;      
    }
    if(this.file.value){
      return fs.writeFileSync(this.path.edge, value);
    }
  },
  get: function(){
    if(this.file.value){
      return fs.readFileSync(this.path.edge, 'utf8').trim();
    }
  },
  enumerable: false,
  configurable: false
});

Object.defineProperty(GpioExport.prototype, 'direction', {
  set: function(value){
    if(this.file.value){
      return fs.writeFileSync(this.path.direction, value);
    }
  },
  get: function(){
    if(this.file.value){
      return fs.readFileSync(this.path.direction, 'utf8').trim();
    }
  },
  configurable: false
}); 

Object.defineProperty(GpioExport.prototype, 'activeLow', {
  set: function(value){
    if(this.file.value){
      return fs.writeFileSync(0|this.path.activeLow, value);
    }
  },
  get: function(){
    if(this.file.value){
      return !!(0|fs.readFileSync(this.path.activeLow, 'utf8').trim());
    }
  },
  configurable: false
}); 

module.exports = {
  Gpio: GpioExport
};
